/*
 * vmfs-tools - Tools to access VMFS filesystems
 * Copyright (C) 2009 Christophe Fillot <cf@utc.fr>
 * Copyright (C) 2009 Mike Hommey <mh@glandium.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Type describing a command */
struct _cmd {
   int argc;
   char **argv;
   int piped;
   int append;
   char *redir;
   char *buf;
};

typedef struct _cmd cmd_t;

/* Return a line after having prompted for it */
char *local_readline(const char *prompt);

/* Return a command after having prompted for it */
const cmd_t *readcmd(const char *prompt);

/* Free a command */
void freecmd(const cmd_t *cmd);
